unit ApeTag;
// Copyright (c) 2002 Erik Stenborg (erik.stenborg@home.se)

{$define utf}
interface

type

  TApeItemType = (TagText, TagBinary);

  TApeTagItem = record
    Key: String;
    Data: WideString;
    ReadOnly: Boolean;
    DataType: TApeItemType;
  end;

  TApeTagItems = array of TApeTagItem;

  TApeTag = Class(TObject)
    private
      tags: TApeTagItems;
      tagatbegin: Boolean;
      tagatend: Boolean;
      FExists: Boolean;
      FVersion: Single;
      function GetKey(Index: Integer): String;
      procedure PutKey(Index: Integer; const Key: String);
      function GetValueI(Index: Integer): Variant;
      procedure PutValueI(Index: Integer; const Value: Variant);
      function GetTitle: Variant;
      procedure PutTitle(Value: Variant);
      function GetArtist: Variant;
      procedure PutArtist(Value: Variant);
      function GetAlbum: Variant;
      procedure PutAlbum(Value: Variant);
      function GetTrack: Variant;
      procedure PutTrack(Value: Variant);
      function GetDate: Variant;
      procedure PutDate(Value: Variant);
      function GetYear: Variant;
      procedure PutYear(Value: Variant);
      function GetComment: Variant;
      procedure PutComment(Value: Variant);
      function GetGenre: Variant;
      procedure PutGenre(Value: Variant);
      function GetIndex(Key: String): Integer;
      function FGetCount: Integer;
      procedure FSetCount(n: Integer);
      procedure FSetVersion(n: Single);
      procedure FSaveToFile(const FileName: string);
    public
      constructor Create;
      procedure ReadFromFile(const FileName: string);
      procedure RemoveFromFile(const FileName: string);
      procedure SaveToFile(const FileName: string);

      property Exists: Boolean read FExists;
      property Version: Single read FVersion write FSetVersion;

      procedure ResetData;
      function GetValue(const Key: String): Variant;
      procedure SetValue(const Key: String; const Value: Variant);
      procedure Add(const Key: String; const Value: Variant);
      procedure Delete(const Index: Integer);

      property Count: Integer read FGetCount write FSetCount;
      property Keys[Index: Integer]: String read GetKey write PutKey;
      property Values[Index: Integer]: Variant read GetValueI write PutValueI;
      property Title: Variant read GetTitle write PutTitle;
      property Artist: Variant read GetArtist write PutArtist;
      property Album: Variant read GetAlbum write PutAlbum;
      property Track: Variant read GetTrack write PutTrack;
      property Date: Variant read GetDate write PutDate;
      property Year: Variant read GetYear write PutYear;
      property Comment: Variant read GetComment write PutComment;
      property Genre: Variant read GetGenre write PutGenre;
  end;

implementation

uses
{$ifdef utf}
//  JclUnicode,
  UTF8Lib,
{$endif}
  Sysutils;

const
  APETAGEX = 'APETAGEX';

type
  TApeTagHead = record
    Apetagex: Array [1..8] of Char;
    Version, TagSize, ItemCount, Flags: Integer;
    Reserved: Array [1..8] of Char;
  end;

constructor TApetag.Create;
begin
  inherited;
  ResetData;
end;

procedure ReadATH(var f: file; var ath: TApeTagHead);
begin
  BlockRead(f, ath, SizeOf(ath));
end;

function CheckApe(ath: TApeTagHead): Boolean;
var
  i: Integer;
begin
  Result := True;
  for i := 1 to 8 do
    if (ath.Apetagex[i] <> APETAGEX[i]) or (ath.Reserved[i] <> Chr(0)) then begin
      Result := False;
      Exit;
    end;
end;

procedure TApetag.ReadFromFile(const FileName: string);
  procedure ReadATI(var f: File; var ati: TApeTagItem; oldstyle: boolean);
  var
    flags, size: Integer;
    ch: Char;
    p: PChar;
  begin
    BlockRead(f, size, SizeOf(size));
    BlockRead(f, flags, SizeOf(flags));
    if (flags and 2) > 0 then ati.DataType := TagBinary
    else ati.DataType := TagText;
    ati.ReadOnly := (flags and 1) > 0;
    BlockRead(f, ch, 1);
    ati.Key := '';
    while (ch<>Chr(0)) do begin
      ati.Key := ati.Key + ch;
      BlockRead(f, ch, 1);
    end;
    GetMem(p, size+1);  //Need termination of the string
    (p+size)^ := Chr(0);
    BlockRead(f, p^, size);
    if (ati.DataType = TagBinary) or oldstyle then
      ati.Data := String(p)
    else
{$ifdef utf}
      ati.Data := UTF8ToWideString(String(p));
{$else}
      ati.Data := String(p);
{$endif}
    FreeMem(p, size+1);
  end;

var
  f: file;
  ath, ath1, ath2: TApeTagHead;
  i: Integer;
  c1, c2: Boolean;
begin
  ResetData;
  AssignFile(f, FileName);
  Reset(f, 1);
  if FileSize(f) < SizeOf(ath) then begin
    c1 := false;
    c2 := false;
  end else begin
    ReadATH(f, ath1);
    c1 := checkape(ath1);
    Seek(f, FileSize(f)-SizeOf(ath));
    ReadATH(f, ath2);
    c2 := checkape(ath2);
  end;
  if not c1 and not c2 then begin
    FExists := False;
    Exit;
  end else if c1 and not c2 then begin
    FExists := True;
    tagatend := false;
    tagatbegin := true;
    ath := ath1;
  end else if not c1 and c2 then begin
    FExists := True;
    tagatend := true;
    tagatbegin := false;
    ath := ath2;
  end else if c1 and c2 then begin
    FExists := True;
    tagatend := true;
    tagatbegin := true;
//    if ath1 <> ath2 then raise Exception.Create('Two tag error');
    ath := ath2;
  end else begin
    FExists := False;
    Exit;
  end;
  FVersion := 0.001 * ath.Version;
  SetLength(tags, ath.ItemCount);
  if (ath.Flags and (1 shl 29)) = 0 then //tag header is at the end of the tag
    Seek(f, FilePos(f)-ath.TagSize);
  for i := 0 to ath.ItemCount-1 do
    ReadATI(f, tags[i], (ath.Version <= 1000));

  CloseFile(f);
end;

procedure TApetag.RemoveFromFile(const FileName: string);
begin
  tagatbegin := false;
  tagatend := false;
  FSaveToFile(FileName);
end;

procedure TApetag.SaveToFile(const FileName: string);
begin
  tagatend := true;
  if FVersion = 0.0 then
    FVersion := 1.0;
  FSaveToFile(FileName);
end;


procedure WriteATI(var f: File; ati: TApeTagItem; oldstyle: Boolean);
var
  b1, b2: PChar;
  n, flags: Integer;
  s: String;
begin
  flags := 0;
  if oldstyle then begin
    s := ati.Data;
    b2 := PChar(s);
    n := Length(s) +1;
  end else begin // from version 2
    //ati.Key := UpperCase(ati.Key);
    for n := 1 to Length(ati.Key) do
      if not (UpCase(ati.Key[n]) in
                            ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
                             'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
                             'U', 'V', 'W', 'X', 'Y', 'Z', '/', '.', '_'])
      then ati.Key[n] := '_';
    if ati.DataType = TagBinary then begin
      flags := flags or 2;
      s := ati.Data;
    end else
{$ifdef utf}
    s := WideStringToUTF8(ati.Data);
{$else}
    s := ati.Data;
{$endif}
    b2 := PChar(s);
    n := Length(s);
    if ati.ReadOnly then flags := flags or 1;
  end;
  b1 := PChar(ati.Key);
  BlockWrite(f, n, SizeOf(n));
  BlockWrite(f, flags, SizeOf(flags));
  BlockWrite(f, b1^, Length(ati.Key)+1);
  BlockWrite(f, b2^, n);
end;

procedure WriteATH(var f: File; ath: TApeTagHead);
begin
  BlockWrite(f, ath, SizeOf(ath));
end;

procedure WriteTagHere(var f: File; var ath: TApeTagHead; atis: TApeTagItems);
var
  i, pos1, pos2: Integer;
begin
  pos1 := FilePos(f);
  if ath.Version >= 2000 then begin
    ath.Flags := ath.Flags or (1 shl 31);
    Seek(f, pos1 + SizeOf(ath));
  end;
  pos2 := FilePos(f);
  for i := 0 to ath.ItemCount-1 do
    if (atis[i].Data <> '') then WriteATI(f, atis[i], (ath.Version <= 1000));
  ath.TagSize := FilePos(f) - pos2 + SizeOf(ath);
  WriteATH(f, ath);

  if ath.Version >= 2000 then begin
    Seek(f, pos1);
    ath.Flags := ath.Flags or (1 shl 29);
    WriteATH(f, ath);
    Seek(f, pos2 + ath.TagSize);
  end;
end;


procedure TApetag.FSaveToFile(const FileName: string);
  procedure SkipTag(var f: File);
  var
    n: Integer;
    ath: TApeTagHead;
  begin
    n := FilePos(f);
    if (FileSize(f) < SizeOf(ath)) then
      Exit
    else
      ReadATH(f, ath);
    if checkape(ath) then
      Seek(f, SizeOf(ath) + ath.TagSize)
    else
      Seek(f, n);
  end;

var
  version: Integer;
  ath, oldath: TApeTagHead;
  i: Integer;
  f, g: file;
  buf: array [1..1024] of Byte;
begin
  for i := 1 to 8 do begin
    ath.Apetagex[i] := APETAGEX[i];
    ath.Reserved[i] := Chr(0);
  end;
  version := Round(FVersion*1000);
  ath.Version := Version;
  ath.ItemCount := Length(tags);
  ath.Flags := 0;

  AssignFile(f, FileName);
  Reset(f,1);
  if FileSize(f) < SizeOf(ath) then
    oldath.Apetagex[1] := Chr(0) //make sure it's invalid
  else
    ReadATH(f, oldath);
  // if we want to write to the beginning
  // or if there is already a tag there
  if tagatbegin or checkape(oldath) then begin
    //rename old file and create new one
    CloseFile(f);
    RenameFile(FileName, FileName+'.old');
    AssignFile(f, FileName);
    Rewrite(f,1);
    //write tag at beginning
    if tagatbegin then
      WriteTagHere(f, ath, tags);
    //append old file
    AssignFile(g, FileName+'.old');
    Reset(g,1);
    SkipTag(g);
    while not EOF(g) do begin
      BlockRead(g, buf, SizeOf(buf), i);
      BlockWrite(f, buf, i);
    end;
    CloseFile(g);
    DeleteFile(PChar(FileName+'.old'));
  end;

  //find beginning of old tag at end of file, and overwrite it
  if (FileSize(f) < SizeOf(ath)) then begin
    Seek(f, FileSize(f));
  end else begin
    Seek(f, FileSize(f)-SizeOf(ath));
    ReadATH(f, oldath);
    if checkape(oldath) then
      if ((oldath.Flags and (1 shl 31)) <> 0) then
        //Old tag with header and footer
        Seek(f, FileSize(f)-SizeOf(ath)-oldath.TagSize)
      else
        //Old tag with only footer
        Seek(f, FileSize(f)-oldath.TagSize);
  end;

  if tagatend then
    WriteTagHere(f, ath, tags);
  Truncate(f);
  CloseFile(f);
end;

procedure TApetag.ResetData;
begin
//  SetLength(tags, 0);
  tags := nil;
  FVersion := 0;
  FExists := false;
end;

function TApetag.GetValue(const Key: String): Variant;
var
  n: Integer;
begin
  n := GetIndex(Key);
  if n <> -1 then
    Result := GetValueI(n);
  //else
    //which one is better? null, empty string or unassigned?
    //Result := Null;
    //Result := '';
end;

procedure TApetag.SetValue(const Key: String; const Value: Variant);
var
  n: Integer;
begin
  n := GetIndex(Key);
  if n <> -1 then
    PutValueI(n, Value)
  else
    Add(Key, Value);
end;

procedure TApetag.Add(const Key: String; const Value: Variant);
var
  n: Integer;
begin
  n := Length(tags);
  SetLength(tags, n+1);
  PutKey(n, Key);
  PutValueI(n, Value);
end;

procedure TApetag.Delete(const Index: Integer);
var
  i,n: Integer;
begin
  n := Length(tags);
  for i := Index to n-2 do
    tags[i] := tags[i+1];
  SetLength(tags, n-1);
end;

function TApetag.FGetCount: Integer;
begin
  Result := Length(tags);
end;

procedure TApetag.FSetCount(n: Integer);
begin
  SetLength(tags, n);
end;

procedure TApetag.FSetVersion(n: Single);
begin
  if n = 1.0 then
    FVersion := 1.0
  else if n = 2.0 then
    FVersion := 2.0;
end;

function TApetag.GetIndex(Key: String): Integer;
var
  i: Integer;
begin
  Result := -1;
  for i := 0 to Length(tags)-1 do
    if UpperCase(tags[i].Key) = UpperCase(Key) then begin
      Result := i;
      Exit;
    end;
end;

function TApetag.GetKey(Index: Integer): String;
begin
  Result := tags[index].Key;
end;

procedure TApetag.PutKey(Index: Integer; const Key: String);
begin
  tags[index].Key := Key;
end;

function TApetag.GetValueI(Index: Integer): Variant;
  function StringToVarByteArray(s: String): Variant;
  var
    n, j: Integer;
  begin
    n := Length(s);
    Result := VarArrayCreate([0, n-1], varByte); // or some other type?
    for j := 0 to n-1 do
      Result[j] := s[j+1];
  end;

begin
  if FVersion <= 1.0 then
    Result := tags[index].Data
  else
    if tags[index].DataType = TagText then
      Result := tags[index].Data
    else
      Result := StringToVarByteArray(tags[index].Data);
end;

procedure TApetag.PutValueI(Index: Integer; const Value: Variant);
  function VarByteArrayToString(v: Variant): String;
  var
    j, n: Integer;
  begin
    n := VarArrayHighBound(v, 1);
    SetLength(Result, n+1);
    for j := VarArrayLowBound(v, 1) to n do
      Result[j+1] := Chr(Byte(v[j]));
  end;

begin
  if tags[Index].ReadOnly then Exit;
  if VarIsArray(Value) then begin
    tags[Index].DataType := TagBinary;
    tags[Index].Data := VarByteArrayToString(Value);
//  end else if VarType(Value) = varDispatch then begin
  end else begin
    tags[Index].DataType := TagText;
    tags[Index].Data := Value;
  end;
end;

function TApetag.GetTitle: Variant;
begin
  Result := GetValue('Title');
end;

procedure TApetag.PutTitle(Value: Variant);
begin
  SetValue('Title', Value);
end;

function TApetag.GetArtist: Variant;
begin
  Result := GetValue('Artist');
end;

procedure TApetag.PutArtist(Value: Variant);
begin
  SetValue('Artist', Value);
end;

function TApetag.GetAlbum: Variant;
begin
  Result := GetValue('Album');
end;

procedure TApetag.PutAlbum(Value: Variant);
begin
  SetValue('Album', Value);
end;

function TApetag.GetTrack: Variant;
begin
  Result := GetValue('Track');
end;

procedure TApetag.PutTrack(Value: Variant);
begin
  SetValue('Track', Value);
end;

function TApetag.GetDate: Variant;
begin
  Result := GetValue('Date');
end;

procedure TApetag.PutDate(Value: Variant);
begin
  SetValue('Date', Value);
end;

function TApetag.GetYear: Variant;
begin
  Result := GetValue('Year');
end;

procedure TApetag.PutYear(Value: Variant);
begin
  SetValue('Year', Value);
end;

function TApetag.GetComment: Variant;
begin
  Result := GetValue('Comment');
end;

procedure TApetag.PutComment(Value: Variant);
begin
  SetValue('Comment', Value);
end;

function TApetag.GetGenre: Variant;
begin
  Result := GetValue('Genre');
end;

procedure TApetag.PutGenre(Value: Variant);
begin
  SetValue('Genre', Value);
end;

end.

